%
%                           F O U R I E R
%
%                           by John Walker
%                      http://www.fourmilab.ch/
%
%   What's all this, you ask?  Well, this is a "literate program",
%   written in the CWEB language created by Donald E. Knuth and
%   Silvio Levy.  This file includes both the C source code for
%   the program and internal documentation in TeX.  Processing
%   this file with the CTANGLE utility produces the C source file,
%   while the CWEAVE program emits documentation in TeX.  The
%   current version of these programs may be downloaded from:
%
%       http://www-cs-faculty.stanford.edu/~knuth/cweb.html
%
%   where you will find additional information on literate
%   programming and examples of other programs written in this
%   manner.
%
%   If you don't want to wade through all these details, don't
%   worry; this distribution includes a .c file already
%   extracted and ready to compile.  If "make" complains that it
%   can't find "ctangle" or "cweave", just "touch *.c"
%   and re-make--apparently the process of extracting the files
%   from the archive messed up the date and time, misleading
%   make into believing it needed to rebuild those files.

@** Introduction.

\vskip 15pt
\centerline{\ttitlefont FOURIER}
\vskip 10pt
\centerline{\titlefont Fourier Transform Utilities}
\vskip 15pt
%\centerline{by John Walker}
%\centerline{\.{http://www.fourmilab.ch/}}
\centerline{\pdfURL{John Walker}{http://www.fourmilab.ch/}}

\vskip 15pt
\centerline{This program is in the public domain.}

\vskip 15pt

This program provides a general purpose fast Fourier transform,
its inverse, and computation of a power spectrum from a
Fourier transform.

\vskip 30pt

@(fourier_test.c@>=
#define REVDATE "5th February 2002"

@** Program global context.

@q Define standard C++ namespaces and classes @>
@i "cweb/c++lib.w"

@c
#include "config.h"                   /* System-dependent configuration */

@h

@<Application include files@>@/
@<Class implementations@>@/

@
We export the class definitions for this package in the
external file \.{fourier.h} that programs which use this
library may include.

@(fourier.h@>=

#ifndef FOURIER_HEADER_DEFINES
#define FOURIER_HEADER_DEFINES
#include <math.h>       // Make sure \.{math.h} is available

#include <iostream>
#include <exception>
#include <stdexcept>
#include <string>
#include <vector>
#include <algorithm>
using namespace std;@#

#include <assert.h>

@<Class definitions@>
#endif

@
The following include files provide access to external components
of the program not defined herein.

@<Application include files@>=
#include "fourier.h"       // Class definitions for this package

@** Fourier transform.

The |fourierTransform| class provides a general-purpose fast Fourier
transform, both forward and inverse.  It operates on arrays of
complex numbers (type |fourierTransform::complex|) which must be a
power of two in length.  Once the transform is initialised for
a given data vector size, any number of vectors of that size
may be transformed.

@
Begin by declaring a table of powers of two which will come handy
in the transformation code.

@<Class definitions@>=
    static const int pwr_two[] = {
        	1,           2,            4,           8,
               16,          32,           64,         128,
              256,         512,         1024,        2048,
             4096,        8192,        16384,       32768,
            65536,      131072,       262144,      524288,
	  1048576,     2097152,      4194304,     8388608,
	 16777216,    33554432,     67108864,   134217728,
	268435456,   536870912,   1073741824
    };
    

@
We begin the declaration of the |fourierTransform| by declaring its
|complex| type, which users of the class refer to as
|fourierTransform::complex|.

@<Class definitions@>=
class fourierTransform {
public:@/
    typedef struct {
	double r, i;
    } complex;

@
Private instance variables hold the $\sin$ and  $\cos$ tables we
use during the transforms and the size of the data vector given as
$\ln_2n$ where $n$ is the number of |complex| numbers in the
vector.

@<Class definitions@>=
private:@/
    
    complex *c;
    int ln;
    
    void buildtable(void);
@
We provide the following query functions to return properties of the transform.

@<Class definitions@>=
    unsigned int getTransformSize(void) {
    	return pwr_two[ln];
    }
    
    unsigned int getTransformSizePowerOfTwo(void) {
    	return static_cast<unsigned int>(ln);
    }

@
If you know the size of the data table in advance, you may simply construct
the |fourierTransform| with its size as the argument.  Otherwise, you can
construct the transform and then set the size later with |setTransformSize|.
You can change the transform size at any time by calling |setTransformSize|.
The sine and cosine table is automatically released by the destructor.
    
@<Class definitions@>=
public:@/
    fourierTransform(unsigned int nsamples) : c(NULL) {
    	setTransformSize(nsamples);
    }

    fourierTransform() : c(NULL) {
    }

    void setTransformSize(unsigned int nsamples) {
    	int p = 1;
	
	for (ln = 0; p < nsamples; ln++) {
	    p <<= 1;
	}
	if (pwr_two[ln] != nsamples) {
	    throw(out_of_range("fourierTransform::setTransformSize: data vector size is not a power of two"));
	}
	buildtable();
    }
    
    ~fourierTransform() {
    	if (c != NULL) {
	    delete c;
	}
    }

@
We wind up the class definitions by declaring the functions which actually perform
the transformation.  The low-level |bitreverse|, |fft|, and |ifft| methods
leave the bit reversing up to the caller, which can save time if you're
twiddling things at low level.  The high-level |transform| and |inverse_transform|
methods do the bit reversing internally.
    
@<Class definitions@>=
    void bitreverse(complex *x);
    void fft(complex *x);
    void ifft(complex *x);
    
    void transform(complex *x) {
    	bitreverse(x);
	fft(x);
    }
    
    void inverse_transform(complex *x) {
    	bitreverse(x);
	ifft(x);
    }
    
    void power_spectrum(const complex *x, double *p);
};

@
The private |buildtable| method generates the sine and cosine tables used
by the transformation.  These are kept in the real and imaginary parts
of an array of |complex| half the size of the data table.

@<Class implementations@>=
void fourierTransform::buildtable(void) {
    int n, halfn, i;
    double theta;
    static const double PI2 = 6.283185307179586476925287;

    n = pwr_two[ln];
    halfn = pwr_two[ln - 1];
    if (c != NULL) {
        delete c;
    }
    c = new complex[halfn];
    for (i = 0; i < halfn; i++) {
        theta = (PI2 * i) / n;
        c[i].r =  cos(theta);
        c[i].i = -sin(theta);
    }
}

@
If you're using the low level |fft| and |ifft| methods, you'll need to call
|bitreverse| to re-order the input vector before calling them.

@<Class implementations@>=
void fourierTransform::bitreverse(complex *x) {
    int i, j, k, n;
    complex hold;

    n = pwr_two[ln];
    k = 0;
    j = 0;
    while (true) {
        if (k > j) {
            hold.r = x[j].r;
            hold.i = x[j].i;

            x[j].r = x[k].r;
            x[j].i = x[k].i;

            x[k].r = hold.r;
            x[k].i = hold.i;
        }
        j++;
        if (j == n) {
            break;
        }

    	@#@,
        /* |k = k + 1|, (|k| is |n| bits long and bit reversed). */

        i = ln - 1;
        while (i >= 0) {
            if (k & pwr_two[i]) {
                k = k - pwr_two[i];
            } else {
                break;
            }
            i--;
        }
        k = k + pwr_two[i];
    }
}

@
Fourier transform.  You must call |bitreverse| first, or use the
high level |transform| method which does this for you.

@<Class implementations@>=
void fourierTransform::fft(complex *x) {
    unsigned int stuckbit, stuckbitcomplement, halfn, t, b;
    unsigned int shft, msk, wt;
    unsigned int c0, c1;
    unsigned int n;
    double qr, qi, xtr, xti, xbr, xbi, cwtr, cwti;

    assert(c != NULL);
    n = pwr_two[ln];
    msk = n - 1;
    halfn = pwr_two[ln - 1];
    shft = ln - 1;

    stuckbit = 1;
    c0 = ln;
    while (c0--) {
        t = 0;
        stuckbitcomplement = ~stuckbit;
        c1 = halfn;
        while (c1--) {
            wt = (t << shft) & msk;
            b = t + stuckbit;

            xbr = x[b].r;
            xbi = x[b].i;
            cwtr = c[wt].r;
            cwti = c[wt].i;

            qr = xbr * cwtr - xbi * cwti;
            qi = xbr * cwti + xbi * cwtr;

            xtr = x[t].r;
            xti = x[t].i;

            x[b].r = xtr-qr;
            x[b].i = xti-qi;

            x[t].r = xtr + qr;
            x[t].i = xti + qi;

            t = (b + 1) & stuckbitcomplement;
        }
        stuckbit <<= 1;
        shft -= 1;
    }
}

@
Inverse Fourier transform.  You must call |bitreverse| first, or use the
high level |inverse_transform| method which does this for you.  Note that
values in the inverse transform will be scaled by the size of the
data set and should be divided by its size to recover
the values of the original transform.

@<Class implementations@>=
void fourierTransform::ifft(complex *x)
{
    unsigned int stuckbit, stuckbitcomplement, halfn, t, b;
    unsigned int shft, msk, wt;
    unsigned int c0, c1;
    unsigned int n;
    double qr, qi;

    assert(c != NULL);
    n = pwr_two[ln];
    msk = n - 1;
    halfn = pwr_two[ln - 1];
    shft = ln - 1;

    stuckbit = 1;
    c0 = ln;
    while (c0--) {
        t = 0;
        stuckbitcomplement = ~stuckbit;
        c1 = halfn;
        while (c1--){
            b = t + stuckbit;
            wt = (t << shft) & msk;

            qr =  x[b].r * c[wt].r  + x[b].i * c[wt].i;
            qi = -(x[b].r * c[wt].i) + x[b].i * c[wt].r;

            x[b].r = x[t].r-qr;
            x[b].i = x[t].i-qi;

            x[t].r += qr;
            x[t].i += qi;

            t = (b + 1) & stuckbitcomplement;
        }
        stuckbit <<= 1;
        shft -= 1;
    }
}

@
Compute the power spectrum from a Fourier transform of a data set.
The power spectrum is just the the absolute values of the
complex Fourier transform vector.

@<Class implementations@>=
void fourierTransform::power_spectrum(const complex *x, double *p)
{
    unsigned int j;
    
    for (j = 0; j < pwr_two[ln]; j++) {
#define sqr(x) ((x) * (x))
    	p[j] = sqrt((sqr(x[j].r) + sqr(x[j].i)) / pwr_two[ln]);
#undef sqr
    }
}

@** Test program.

@(fourier_test.c@>=

@<Test program include files@>;

@<Show how to call test program@>;

#define FFTPOW  16
#define FFTSIZE (1 << FFTPOW)

int main(int argc, char *argv[])
{
    extern char *optarg;            /* Imported from |getopt| */
    extern int optind;

    int opt;
    
try {

@<Process command-line options@>;@/

    fourierTransform ft(FFTSIZE);
    
    fourierTransform::complex a[FFTSIZE], b[FFTSIZE], c[FFTSIZE];
    
    /*	Synthesise a test function from superposed sine curves.  */
#define P1 773.0
#define A1 1
#define P2 593.0
#define A2 0.5
#define P3 191.0
#define A3 1
    for (int i = 0; i < FFTSIZE; i++) {
    	a[i].r = sin(static_cast<double>(i) / P1) * A1;
#ifdef P2
	a[i].r += sin(static_cast<double>(i) / P2) * A2;
#endif
#ifdef P3
	a[i].r += sin(static_cast<double>(i) / P3) * A3;
#endif
	a[i].i = 0;
    }
    
#ifdef CRAPTASTROPHE
    b = a;
#else
    memcpy(b, a, sizeof b);
#endif
    ft.transform(b);

#ifdef CRAPTASTROPHE
    c = b;
#else
    memcpy(c, b, sizeof c);
#endif
    ft.inverse_transform(c);
    
    string fileName = "fourier";

    ofstream gp((fileName + ".gp").c_str()),
    	     dat((fileName + ".dat").c_str());
	     
    /* Data columns:
    
    	    	1:  Index
		2:  Input function
		3:  Power spectrum
		4:  Inverse FFT results
    */

    double pspec[FFTSIZE];
    ft.power_spectrum(b, pspec);
#define PLOTSIZE 1000	     
    for (int j = 0; j < PLOTSIZE; j++) {
	c[j].r /= PLOTSIZE;
	c[j].i /= PLOTSIZE;
    	dat << j << " " << a[j].r << " " << pspec[j] << " " << c[j].r << endl;
    }
    
    gp << "set term pbm small color" << endl;
    
#define PLOT_BOTH

#ifdef PLOT_INPUT
    gp << "plot \"" << fileName << ".dat\" using 1:2 with lines" << endl;
#endif

#ifdef PLOT_INVERSE
    gp << "plot \"" << fileName << ".dat\" using 1:4 with lines" << endl;
#endif

#ifdef PLOT_BOTH
    gp << "plot \"" << fileName << ".dat\" using 1:2 with lines, ";
    gp << "\"" << fileName << ".dat\" using 1:4 with lines" << endl;
#endif

#ifdef PLOT_POWER	  
    gp << "plot \"" << fileName << ".dat\" using 1:3 with lines" << endl;
#endif
    
    string command("gnuplot ");
    command += fileName + ".gp | ppmtogif >" + fileName + ".gif";
#ifdef DEV_PLOT_DEBUG
    cout << command << endl;
#else
    command += " 2>/dev/null";
#endif
    gp.close();
    dat.close();
    system(command.c_str());
#ifndef DEV_PLOT_DEBUG
    //	Delete the temporary files used to create the plot
    remove((fileName + ".gp").c_str());
    remove((fileName + ".dat").c_str());
#endif
    
} catch (exception &e) {
    cout << "Blooie!!!  Exception popped: " << e.what() << endl;
#ifndef CORE_DUMP
#ifdef STACK_TRACE
    char s[160];
    sprintf(s, "/bin/echo 'where\nq' >/tmp/gdbcmd ; gdb  -batch --command "
    	"/tmp/gdbcmd %s %d", argv[0], getpid());
    system(s);
    sleep(5);
#endif
    throw;  	    // Re-throw exception to dump core
#endif
}

    return 0;
}

@
We use |getopt| to process command line options.  This
permits aggregation of options without arguments and
both \.{-d}{\it arg} and \.{-d} {\it arg} syntax.
@<Process command-line options@>=
    while ((opt = getopt(argc, argv, "nu-:")) != -1) {
        switch (opt) {

            case 'u':             /* \.{-u}  Print how-to-call information */
            case '?':
                usage();
                return 0;

            case '-':             /* \.{--}  Extended options */
                switch (optarg[0]) {
                    case 'c':     /* \.{--copyright} */
                        cout << "This program is in the public domain.\n";
                        return 0;

                    case 'h':     /* \.{--help} */
                        usage();
                        return 0;

                    case 'v':     /* \.{--version} */
                        cout << PRODUCT << " " << VERSION << "\n";
                        cout << "Last revised: " << REVDATE << "\n";
                        cout << "The latest version is always available\n";
                        cout << "at http://www.fourmilab.ch/eggtools/eggshell\n";
                        return 0;
                }
        }
    }

@ Procedure |usage|
prints how-to-call information.

@<Show how to call test program@>=
static void usage(void)
{
    cout << PRODUCT << "  --  Analyse eggsummary files.  Call:\n";
    cout << "              " << PRODUCT << " [options] [infile] [outfile]\n";
    cout << "\n";
    cout << "Options:\n";
    cout << "           --copyright       Print copyright information\n";
    cout << "           -u, --help        Print this message\n";
    cout << "           --version         Print version number\n";
    cout << "\n";
    cout << "by John Walker\n";
    cout << "http://www.fourmilab.ch/\n";
}

@
We need the following definitions to compile the test program.

@<Test program include files@>=

#include "config.h"         // Our configuration

/* C++ include files  */
#include <iostream>
#include <fstream>
#include <exception>
#include <stdexcept>
#include <string>
using namespace std;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef HAVE_GETOPT
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#else
#include "getopt.h"     /* No system \.{getopt}---use our own */
#endif
#include "fourier.h"        // Class definitions for this package


@** Index.
The following is a cross-reference table for \.{fourier}.
Single-character identifiers are not indexed, nor are
reserved words.  Underlined entries indicate where
an identifier was declared.
