%
%                          P L O T T I N G
%
%                           by John Walker
%                      http://www.fourmilab.ch/
%
%   What's all this, you ask?  Well, this is a "literate program",
%   written in the CWEB language created by Donald E. Knuth and
%   Silvio Levy.  This file includes both the C source code for
%   the program and internal documentation in TeX.  Processing
%   this file with the CTANGLE utility produces the C source file,
%   while the CWEAVE program emits documentation in TeX.  The
%   current version of these programs may be downloaded from:
%
%       http://www-cs-faculty.stanford.edu/~knuth/cweb.html
%
%   where you will find additional information on literate
%   programming and examples of other programs written in this
%   manner.
%
%   If you don't want to wade through all these details, don't
%   worry; this distribution includes a .c file already
%   extracted and ready to compile.  If "make" complains that it
%   can't find "ctangle" or "cweave", just "touch *.c"
%   and re-make--apparently the process of extracting the files
%   from the archive messed up the date and time, misleading
%   make into believing it needed to rebuild those files.

@** Introduction.

\vskip 15pt
\centerline{\ttitlefont PLOTTING}
\vskip 10pt
\centerline{\titlefont Graph Plotting Library}
\vskip 15pt
%\centerline{by John Walker}
%\centerline{\.{http://www.fourmilab.ch/}}
\centerline{\pdfURL{John Walker}{http://www.fourmilab.ch/}}

\vskip 15pt
\centerline{This program is in the public domain.}

\vskip 15pt

{\narrower\smallskip\bf
    This file is purely a placeholder.  It will eventually be the high-level framework
    for making plots without invoking \.{GNUPLOT}.
\smallskip}

\vskip 30pt

@(plotting_test.c@>=
#define REVDATE "4th June 2001"

@** Program global context.

@q Define standard C++ namespaces and classes @>
@i "cweb/c++lib.w"

@c
#include "config.h"                   /* System-dependent configuration */

@h

@<Application include files@>@/
@<Class implementations@>@/

@
We export the class definitions for this package in the
external file \.{plotting.h} that programs which use this
library may include.

@(plotting.h@>=

#ifndef PLOTTING_HEADER_DEFINES
#define PLOTTING_HEADER_DEFINES
#include <math.h>       // Make sure \.{math.h} is available

#include <iostream>
#include <exception>
#include <stdexcept>
#include <string>
#include <vector>
#include <algorithm>
using namespace std;@#

#include "bitmap.h"
#include <stdio.h>      // Needed to define \.{FILE} for \.{gifout.h}
#include "gifout.h"
#include "graphics.h"
#include "pstampr.h"
#include "psrtext.h"

@<Class definitions@>
#endif

@
The following include files provide access to external components
of the program not defined herein.

@<Application include files@>=
#include "plotting.h"       // Class definitions for this package

@
The following classes are defined and their
implementations provided.

@<Class implementations@>=
@<Colour systems@>@/

@** Colour system parent class: csColour.

This is the parent class of all colour classes.

@<Class definitions@>=
class csColour {
public:@/
    virtual string colourSystemName(void) = 0;  // Return colour system name

    virtual void writeParameters(ostream &of) {
        //  Write description of colour system to output stream
        of << "Colour system: " << colourSystemName() << "\n";
    }
};

@*1 Spectral colour systems parent class: csSpectralColour.

The most general of all our notions of colour is that of an function
that maps wavelengths (given in metres) onto intensities.  This form
allows specification of any radiation in the electromagnetic spectrum.
Intensities are normalised to the range zero to one by dividing the
intensity for a given frequency by the integrated intensity over the
entire electromagnetic spectrum.

@<Class definitions@>=
class csSpectralColour : public csColour {
public:@/

    // Return normalised intensity for a given wavelength in metres

    virtual double getIntensity(double waveLength) = 0;

    // Return |true| if this is a pure (monochromatic) radiator, |false|
    // otherwise.

    virtual bool isMonochromatic(void) = 0;
};

@*2 Monochromatic spectral colour systems: csMonochromaticColour.

A monochromatic colour is a abstract notion of a source which
radiates all its intensity at a single wavelength.  These don't
really exist in the real world, since fundamental quantum
processes will always spread the spectrum of any real radiator.

@<Class definitions@>=
class csMonochromaticColour : public csSpectralColour {
private:@/
    double wavelength;

public:@/
    virtual string colourSystemName(void) {
        return "monochromatic spectral";
    }

    double getIntensity(double waveLength)
    {
        return (waveLength == wavelength) ? 1.0 : 0.0;
    }

    bool isMonochromatic(void) {
        return true;
    }

    // Constructors and destructors

    csMonochromaticColour(void)
    {
        wavelength = 0.0;
    }

    csMonochromaticColour(double waveLength)
    {
        wavelength = waveLength;
    }

    // Class-specific methods

    void setWavelength(double waveLength)
    {
        wavelength = waveLength;
    }

    double getWavelength(void)
    {
        return wavelength;
    }

    void writeParameters(ostream &of) {
        csSpectralColour::writeParameters(of);
        of << "    Wavelength = " << getWavelength() << "\n";
    }
};


@*2 Black: csSpectralBlack.
Black is the absence of colour.  We define it as a spectral colour
with zero intensity at any wavelength.

@<Class definitions@>=
class csSpectralBlack : public csSpectralColour {
public:@/
    virtual string colourSystemName(void) {
        return "black spectral";
    }

    double getIntensity(double waveLength)
    {
        return 0.0;
    }

    bool isMonochromatic(void) {
        return false;
    }
};

@*2 White: csSpectralWhite.
A theoretical source of white noise has equal power at all
wavelengths.  Such a source is impossible in reality since the
energy flux would be infinite.

@<Class definitions@>=
class csSpectralWhite : public csSpectralColour {
public:@/
    virtual string colourSystemName(void) {
        return "white spectral";
    }

    double getIntensity(double waveLength)
    {
        return 1.0;
    }

    bool isMonochromatic(void) {
        return false;
    }
};

@*2 Planckian black body: csBlackBody.
Define the colour of a Planckian black body radiator with a given colour
temperature.

@<Class definitions@>=
class csBlackBody : public csSpectralColour {
private:@/
    // Change temperature to class with system conversions
    double temperature;         // Temperature of radiator ($^\circ K$)

public:@/
    double getIntensity(double waveLength)
    {
        return 0.0;
    }

    bool isMonochromatic(void) {
        return false;
    }

    // Constructors and destructors

    csBlackBody(void)
    {
        temperature = 0.0;
    }

    csBlackBody(double temp)
    {
        temperature = temp;
    }

    // Class-specific methods

    void setTemperature(double temp)
    {
        temperature = temp;
    }

    double getTemperature(void)
    {
        return temperature;
    }

    double totalFlux(void)        // Total energy flux in $W/m^2$
    {
        /* Total energy flux in Watts per square metre of a black
           body with temperature T (degrees Kelvin) is given by:

                $$C_s T^4$$

           where $C_s$ is the Stefan-Boltzman constant:

                $$5.67051\times 10^{-8} W/(m^2 K^4)$$

           which, in fundamental terms is:

                $$(\pi^2 k^4) / (60 (h / (2 \pi))^3 c^2)$$ */

        return 5.67051e-8 *
            (temperature * temperature * temperature * temperature);
    }

    double flux(double wl)    // Energy flux at a given wavelength
    {

        /* Energy flux from a black body with temperature $T$ (degrees
           Kelvin), in Watts per square metre at wavelength $\lambda$
           (metres) is given by:

                $$M_e = C_1 \lambda^{-5} (e^{C_2 / (\lambda * T)} - 1)^{-1}$$

           where:
                $$C_1 = 3.74183e-16 W m^2$$
                $$C_2 = 1.4388e-2 m ^\circ K$$
        */

        return 3.74183e-16 / ((wl * wl * wl * wl * wl) *
                (exp(1.4388e-2 / (wl * temperature)) - 1));
    }

    void writeParameters(ostream &of) {
        csSpectralColour::writeParameters(of);
        of << "    Temperature = " << getTemperature() << " K\n";
    }
};

@*1 Device colour systems parent class: csDeviceColour.
This class is the parent of all classes which specify colour in a
device-specific way, assuming a particular set of illuminants
which are summed to form the colour.

@<Class definitions@>=
class csDeviceColour : public csColour {
public:@/

    @<Device colour fundamental methods@>@/
    @<Device colour derived methods@>@/
    @<Device colour system conversion utilities@>@/
};

@
The following three must-implement methods allow retrieving
the colour in any of the three fundamental device colour
spaces: RGB (additive), CMYK (subtractive), or Greyscale.
Typically, a specific device colour class will store the
colour in one of these forms and implement the other
retrieval methods by converting the colour representation
to that form.

@<Device colour fundamental methods@>=
    virtual void asRGB(double &r, double &g, double &b) = 0;
    virtual void asCMYK(double &c, double &m, double &y, double &k) = 0;
    virtual void asGreyScale(double &g) = 0;

@
The following methods allow retrieval of a device colour in
other colour mapping spaces.  The |csDeviceColour| class
implements methods for these functions which provide default
definitions which use the subclass |asRGB| method to obtain
RGB components which are the converted into the requested
colour space.

@<Device colour derived methods@>=
    virtual void asHSV(double &h, double &s, double &v);
    virtual void asHLS(double &h, double &l, double &s);
    virtual void asYIQ(double &y, double &i, double &q);
    virtual void asYUV(double &y, double &u, double &v);
    virtual void asSMPTE(double &y, double &Pb, double &Pr);
    virtual void asCMY(double &c, double &m, double &y);

@
The following |static| methods of the |csDeviceColour| class
provide interconversions of common colour systems using the
convention that colour components $c$ are |double| values
$0\leq c\leq 1$.

@<Device colour system conversion utilities@>=
protected:@/
    static double hlsval(double n1, double n2, double hue);

public:@/
    static void hsv_rgb(double h, double s, double v,
                        double *r, double *g, double *b);  // HSV $\rightarrow$ RGB
    static void rgb_hsv(double r, double g, double b,
                        double *h, double *s, double *v);  // RGB $\rightarrow$ HSV
    static void rgb_hls(double r, double g, double b,
                        double *h, double *l, double *s);  // RGB $\rightarrow$ HLS
    static void hls_rgb(double h, double l, double s,
                        double *r, double *g, double *b);  // HLS $\rightarrow$ RGB
    static void rgb_yiq(double r, double g, double b,
                        double *y, double *i, double *q);  // RGB $\rightarrow$ YIQ
    static void yiq_rgb(double y, double i, double q,
                        double *r, double *g, double *b);  // YIQ $\rightarrow$ RGB
    static void rgb_yuv(double r, double g, double b,
                        double *y, double *u, double *v);  // RGB $\rightarrow$ YUV
    static void yuv_rgb(double y, double u, double v,
                        double *r, double *g, double *b);  // YUV $\rightarrow$ RGB
    static void rgb_smpte_204M(double r, double g, double b,
                        double *y, double *Pb, double *Pr); // RGB $\rightarrow$ SMPTE~204M
    static void smpte_204M_rgb(double y, double Pb, double Pr,
                        double *r, double *g, double *b);  // SMPTE~204M $\rightarrow$ RGB
    static void rgb_cmy(double r, double g, double b,
                        double *c, double *m, double *y);  // RGB $\rightarrow$ CMY
    static void cmy_rgb(double c, double m, double y,
                        double *r, double *g, double *b);  // CMY $\rightarrow$ RGB
    static void cmy_cmyk(double c, double m, double y,
                        double *oc, double *om, double *oy, double *ok); // CMY $\rightarrow$ CMYK
    static void cmyk_cmy(double c, double m, double y, double k,
                        double *oc, double *om, double *oy); // CMYK $\rightarrow$ CMY

@*2 Colour system conversion utilities.
The following |static| methods of the |csDeviceColour| class
implement interconversions of common colour systems using the
convention that colour components $c$ are |double| values
$0\leq c\leq 1$.

@
|HSV_RGB|: Convert HSV colour specification to RGB intensities.
           Hue ($h$) is specified as a |double| value from 0 to 360$^\circ$,
           Saturation ($s$) and Intensity ($v$) as |double|s from 0 to 1.  The
           ($r,g,b$) components are returned as |double|s from 0 to 1.

@<Colour systems@>=
void csDeviceColour::hsv_rgb(double h, double s, double v,
                             double *r, double *g, double *b)
{
    int i;
    double f, p, q, t;

    if (s == 0) {
        *r = *g = *b = v;
    } else {
        if (h == 360.0)
            h = 0;
        h /= 60.0;

        i = (int) h;
        f = h - i;
        p = v * (1.0 - s);
        q = v * (1.0 - (s * f));
        t = v * (1.0 - (s * (1.0 - f)));
        assert(i >= 0 && i <= 5);
        switch (i) {

        case 0:
            *r = v;
            *g = t;
            *b = p;
            break;

        case 1:
            *r = q;
            *g = v;
            *b = p;
            break;

        case 2:
            *r = p;
            *g = v;
            *b = t;
            break;

        case 3:
            *r = p;
            *g = q;
            *b = v;
            break;

        case 4:
            *r = t;
            *g = p;
            *b = v;
            break;

        case 5:
            *r = v;
            *g = p;
            *b = q;
            break;
        }
    }
}

@
|RGB_HSV|: Map $r,g,b$ intensities in the range from 0 to 1 into
            Hue ($h$), Saturation ($s$), and Value ($v$): Hue from 0 to 360$^\circ$,
            Saturation from 0 to 1, and Value from 0 to 1.
            Special case: if Saturation is 0 (it's a grey scale
            tone), Hue is undefined and is returned as -1.

            This follows Foley {\it et al.}, section 13.3.4.

@<Colour systems@>=
void csDeviceColour::rgb_hsv(double r, double g, double b,
                    double *h, double *s, double *v)
{
    double imax = max(r, max(g, b)),
             imin = min(r, min(g, b)),
             rc, gc, bc;

    *v = imax;
    if (imax != 0)
        *s = (imax - imin) / imax;
    else
        *s = 0;

    if (*s == 0) {
        *h = -1;
    } else {
        rc = (imax - r) / (imax - imin);
        gc = (imax - g) / (imax - imin);
        bc = (imax - b) / (imax - imin);
        if (r == imax)
            *h = bc - gc;
        else if (g == imax)
            *h = 2.0 + rc - bc;
        else
            *h = 4.0 + gc - rc;
        *h *= 60.0;
        if (*h < 0.0)
            *h += 360.0;
    }
}

@
|RGB_HLS|: Map $r,g,b$ intensities in the range from 0 to 1 into
           Hue ($h$), Lightness ($l$), and Saturation ($s$):
           Hue from 0 to 360$^\circ$,
           Lightness from 0 to 1, and Saturation from 0 to 1.
           Special case: if Saturation is 0 (it's a grey scale
           tone), Hue is undefined and is returned as -1.

           This follows Foley {\it et al.}, section 13.3.5.

@<Colour systems@>=
void csDeviceColour::rgb_hls(double r, double g, double b,
                    double *h, double *l, double *s)
{
    double imax = max(r, max(g, b)),
           imin = min(r, min(g, b)),
           rc, gc, bc;

    *l = (imax + imin) / 2;

    if (imax == imin) {
        *s = 0;
        *h = -1;
    } else {
        if (*l <= 0.5)
            *s = (imax - imin) / (imax + imin);
        else
            *s = (imax - imin) /
                 (2.0 - imax - imin);

        rc = (imax - r) / (imax - imin);
        gc = (imax - g) / (imax - imin);
        bc = (imax - b) / (imax - imin);
        if (r == imax)
            *h = bc - gc;
        else if (g == imax)
            *h = 2.0 + rc - bc;
        else
            *h = 4.0 + gc - rc;
        *h *= 60.0;
        if (*h < 0)
            *h += 360.0;
    }
}

@
|HLS_RGB|: Convert HLS colour specification to $r,g,b$ intensities.
           Hue ($h$) is specified as a |double| value from 0 to 360$^\circ$;
           Lightness ($l$) and Saturation ($s$) as |double|s from 0 to 1.  The
           RGB components are returned as |double|s from 0 to 1.

@<Colour systems@>=
double csDeviceColour::hlsval(double n1, double n2, double hue)
{
    if (hue > 360.0)
        hue -= 360.0;
    else if (hue < 0.0)
        hue += 360.0;
    if (hue < 60.0) {
        return n1 + ((n2 - n1) * hue) / 60.0;
    } else if (hue < 180.0) {
        return n2;
    } else if (hue < 240.0) {
        return n1 + ((n2 - n1) * (240.0 - hue)) / 60.0;
    } else {
        return n1;
    }
}

void csDeviceColour::hls_rgb(double h, double l, double s,
                    double *r, double *g, double *b)
{
    double m1, m2;

    if (l <= 0.5)
        m2 = l * (1.0 + s);
    else
        m2 = l + s - (l * s);
    m1 = 2 * l - m2;

    if (s == 0) {
        *r = *g = *b = l;
    } else {
        *r = hlsval(m1, m2, h + 120.0);
        *g = hlsval(m1, m2, h);
        *b = hlsval(m1, m2, h - 120.0);
    }
}

@
|RGB_YIQ| Convert RGB colour specification, $r,g,b$ ranging from
             0 to 1, to $Y,I,Q$ colour specification.  YIQ is
             the encoding used in NTSC television.

$$\left[\matrix{Y\cr I\cr Q\cr}\right]=
\left[\matrix{0.2989 & \hfill 0.5866 & \hfill 0.1144\cr
              0.5959 & -0.2741 & -0.3218\cr
              0.2113 & -0.5227  & \hfill 0.3113\cr}\right]\cdot
\left[\matrix{R\cr G\cr B\cr}\right]
$$

@<Colour systems@>=
void csDeviceColour::rgb_yiq(double r, double g, double b,
                    double *y, double *i, double *q)
{
    double ay = (r * 0.2989 + g *  0.5866 + b *  0.1144),
           ai = (r * 0.5959 + g * -0.2741 + b * -0.3218),
           aq = (r * 0.2113 + g * -0.5227 + b *  0.3113);

    *y = ay;
    if (ay == 1.0) {                  /* Prevent round-off on grey scale */
        ai = aq = 0.0;
    }
    *i = ai;
    *q = aq;
}

@
|YIQ_RGB|: Convert YIQ colour specification, $Y,I, Q$ given  as
         |double|s:  $0\leq Y\leq 1$, $-0.6\leq I\leq 0.6$,
         $-0.52\leq Q\leq 0.52$,
         to $R,G,B$ intensities  in  the  range
         from  0 to 1.  The matrix below is the inverse of the
         |RGB_YIQ| matrix above.   YIQ is the  encoding used  in
         NTSC television.

$$\left[\matrix{R\cr G\cr B\cr}\right]=
\left[\matrix{1.0000 & \hfill 0.9562 & \hfill 0.6210\cr
              1.0000 & -0.2717 & -0.6485\cr
              1.0000 & -1.1053  & \hfill 1.7020\cr}\right]\cdot
\left[\matrix{Y\cr I\cr Q\cr}\right]
$$

%                |R|   |1.0000  0.9562  0.6210|   |Y|
%                |G| = |1.0000 -0.2717 -0.6485| . |I|
%                |B|   |1.0000 -1.1053  1.7020|   |Q|

@<Colour systems@>=
void csDeviceColour::yiq_rgb(double y, double i, double q,
                    double *r, double *g, double *b)
{
    double ar = (y + i *   0.9562 + q *  0.6210),
           ag = (y + i *  -0.2717 + q * -0.6485),
           ab = (y + i *  -1.1053 + q *  1.7020);

    *r = max(0.0, min(1.0, ar));
    *g = max(0.0, min(1.0, ag));
    *b = max(0.0, min(1.0, ab));
}

@
|RGB_YUV|: Convert RGB colour specification, $R,G,B$ ranging
                 from 0 to 1, to $Y,U,V$ colour specification.  YUV
                 is the encoding used by PAL television.

$$\left[\matrix{Y\cr U\cr V\cr}\right]=
\left[\matrix{\hfill 0.2989 & \hfill 0.5866 & \hfill 0.1144\cr
              -0.1473 & -0.2717 & \hfill 0.4364\cr
              \hfill 0.6149 & -0.5145  & -0.1004\cr}\right]\cdot
\left[\matrix{R\cr G\cr B\cr}\right]
$$

%                |Y|   | 0.2989  0.5866  0.1144|   |R|
%                |U| = |-0.1473 -0.2891  0.4364| . |G|
%                |V|   | 0.6149 -0.5145 -0.1004|   |B|

@<Colour systems@>=
void csDeviceColour::rgb_yuv(double r, double g, double b,
                    double *y, double *u, double *v)
{
    double ay = (r *  0.2989 + g *  0.5866 + b *  0.1144),
           au = (r * -0.1473 + g * -0.2891 + b *  0.4364),
           av = (r *  0.6149 + g * -0.5145 + b * -0.1004);

    *y = ay;
    if (ay == 1.0) {                  /* Prevent round-off on grey scale */
        au = av = 0.0;
    }
    *u = au;
    *v = av;
}

@
|YUV_RGB|: Convert  YUV  colour  specification, $Y,U,V$ given as
                 |double|s,  to $R,G,B$ intensities in the range from 0 to
                 1.  The matrix below is the inverse  of  the  |rgb_yuv|
                 matrix  above.   YUV  is  the  encoding  used  by PAL
                 television.

$$\left[\matrix{R\cr G\cr B\cr}\right]=
\left[\matrix{1.0000 & \hfill 0.0000 & \hfill 1.1402\cr
              1.0000 & -0.3959 & -0.5810\cr
              1.0000 & \hfill 2.0294  & \hfill 0.0000\cr}\right]\cdot
\left[\matrix{Y\cr U\cr V\cr}\right]
$$

%                |R|   |1.0000  0.0000  1.1402|   |Y|
%                |G| = |1.0000 -0.3959 -0.5810| . |U|
%                |B|   |1.0000  2.0294  0.0000|   |V|

@<Colour systems@>=
void csDeviceColour::yuv_rgb(double y, double u, double v,
                    double *r, double *g, double *b)
{
    double ar = (y + u *   0.0000 + v *  1.1402),
           ag = (y + u *  -0.3959 + v * -0.5810),
           ab = (y + u *   2.0294 + v *  0.0000);

    *r = max(0.0, min(1.0, ar));
    *g = max(0.0, min(1.0, ag));
    *b = max(0.0, min(1.0, ab));
}

@
|RGB_SMPTE_204M|: Convert RGB colour specification, $R,G,B$
                  ranging from 0 to 1, to $Y,P_b,P_r$ colour
                  specification according to the SMPTE
                  204M (1988) specification for HDTV using
                  the SMPTE set of phosphors.

$$\left[\matrix{Y\cr P_b\cr P_r\cr}\right]=
\left[\matrix{\hfill 0.2122 & \hfill 0.7013 & \hfill 0.0865\cr
              -0.1162 & -0.3838 & \hfill 0.5000\cr
              \hfill 0.6149 & -0.4451  & -0.0549\cr}\right]\cdot
\left[\matrix{R\cr G\cr B\cr}\right]
$$

%                |Y |   | 0.2122  0.7013  0.0865|   |R|
%                |Pb| = |-0.1162 -0.3838  0.5000| . |G|
%                |Pr|   | 0.5000 -0.4451 -0.0549|   |B|

@<Colour systems@>=
void csDeviceColour::rgb_smpte_204M(double r, double g, double b,
                           double *y, double *Pb, double *Pr)
{
    double ay  = (r *  0.2122 + g *  0.7013 + b *  0.0865),
           aPb = (r * -0.1162 + g * -0.3838 + b *  0.5000),
           aPr = (r *  0.5000 + g * -0.4451 + b * -0.0549);

    *y = ay;
    if (ay == 1.0) {                  /* Prevent round-off on grey scale */
        aPb = aPr = 0.0;
    }
    *Pb = aPb;
    *Pr = aPr;
}

@
|SMPTE_204M_RGB|: Convert a colour specified using the SMPTE
                  reference phosphors as given in the SMPTE 204M
                  (1988) specification for HDTV to $R,G,B$
                  intensities in the range from 0 to 1.  The
                  matrix below is the inverse of the
                  |RGB_SMPTE_204M| matrix above.

$$\left[\matrix{R\cr G\cr B\cr}\right]=
\left[\matrix{1.0000 & \hfill 0.0000 & \hfill 1.5755\cr
              1.0000 & -0.2254 & -0.4768\cr
              1.0000 & \hfill 1.8270  & \hfill 0.0000\cr}\right]\cdot
\left[\matrix{Y\cr P_b\cr P_r\cr}\right]
$$

%                |R|   |1.0000  0.0000  1.5755|   |Y |
%                |G| = |1.0000 -0.2254 -0.4768| . |Pb|
%                |B|   |1.0000  1.8270  0.0000|   |Pr|

@<Colour systems@>=
void csDeviceColour::smpte_204M_rgb(double y, double Pb, double Pr,
                           double *r, double *g, double *b)
{
    double ar = (y + Pb *   0.0000 + Pr *  1.5755),
           ag = (y + Pb *  -0.2254 + Pr * -0.4768),
           ab = (y + Pb *   1.8270 + Pr *  0.0000);

    *r = max(0.0, min(1.0, ar));
    *g = max(0.0, min(1.0, ag));
    *b = max(0.0, min(1.0, ab));
}

@
|RGB_CMY|: Convert RGB colour specification, $R,G,B$ ranging
           from 0 to 1, to $C,M,Y$ colour specification, also
           ranging from 0 to 1.

$$\left[\matrix{C\cr M\cr Y\cr}\right]=
\left[\matrix{1\cr 1\cr 1\cr}\right]-
\left[\matrix{R\cr G\cr B\cr}\right]
$$

%                |C|   |1|   |R|
%                |M| = |1| - |G|
%                |Y|   |1|   |B|

@<Colour systems@>=
void csDeviceColour::rgb_cmy(double r, double g, double b,
                    double *c, double *m, double *y)
{
    *c = 1.0 - r;
    *m = 1.0 - g;
    *y = 1.0 - b;
}

@
|CMY_RGB|: Convert CMY colour specification, $C,M,Y$ ranging
           from 0 to 1, to $R,G,B$ colour specification, also
           ranging from 0 to 1.

$$\left[\matrix{R\cr G\cr B\cr}\right]=
\left[\matrix{1\cr 1\cr 1\cr}\right]-
\left[\matrix{C\cr M\cr Y\cr}\right]
$$

%                |R|   |1|   |C|
%                |G| = |1| - |M|
%                |B|   |1|   |Y|

@<Colour systems@>=
void csDeviceColour::cmy_rgb(double c, double m, double y,
                    double *r, double *g, double *b)
{
    *r = 1.0 - c;
    *g = 1.0 - m;
    *b = 1.0 - y;
}

@
|cmy_cmyk|: Convert CMY colour specification, $C,M,Y$ ranging
            from 0 to 1, to $C,M,Y,K$ colour specification, also
            ranging from 0 to 1.  $K$ is the black ink component
            in four colour printing processes.  We convert
            $C,M,Y$ by computing $K=\min(C,M,Y)$, then
            subtracting that value from each of the $C,M,Y$
            components.

@<Colour systems@>=
void csDeviceColour::cmy_cmyk(double c, double m, double y,
                    double *oc, double *om, double *oy, double *ok)
{
    double k = min(c, min(m, y));

    *oc = c - k;
    *om = m - k;
    *oy = y - k;
    *ok = k;
}

@
|cmyk_cmy|: Convert CMYK colour specification, $C,M,Y,K$ ranging
            from 0 to 1, to $C,M,Y$ colour specification, also
            ranging from 0 to 1.  $K$ is the black ink component
            in four colour printing processes.  We simply add
            the $K$ component to each of the $C,M,Y$ components.

@<Colour systems@>=
void csDeviceColour::cmyk_cmy(double c, double m, double y, double k,
                    double *oc, double *om, double *oy)
{
    *oc = c + k;
    *om = m + k;
    *oy = y + k;
}

@*2 Colour system conversion methods.

The following methods express a colour in a given colour system
in a variety of others.  The class must implement the |asRGB|
method required for each of these conversions.

@<Colour systems@>=
void csDeviceColour::asHSV(double &h, double &s, double &v)  // To HSV
{
    double r, g, b;

    asRGB(r, g, b);
    rgb_hsv(r, g, b, &h, &s, &v);
}

void csDeviceColour::asHLS(double &h, double &l, double &s)  // To HLS
{
    double r, g, b;

    asRGB(r, g, b);
    rgb_hls(r, g, b, &h, &l, &s);
}

void csDeviceColour::asYIQ(double &y, double &i, double &q)  // To YIQ
{
    double r, g, b;

    asRGB(r, g, b);
    rgb_yiq(r, g, b, &y, &i, &q);
}

void csDeviceColour::asYUV(double &y, double &u, double &v)  // To YUV
{
    double r, g, b;

    asRGB(r, g, b);
    rgb_yuv(r, g, b, &y, &u, &v);
}

void csDeviceColour::asSMPTE(double &y, double &Pb, double &Pr)  // To SMPTE 204M
{
    double r, g, b;

    asRGB(r, g, b);
    rgb_smpte_204M(r, g, b, &y, &Pb, &Pr);
}

void csDeviceColour::asCMY(double &c, double &m, double &y)  // To CMY
{
    double r, g, b;

    asRGB(r, g, b);
    rgb_cmy(r, g, b, &c, &m, &y);
}

@** Test program.

@(plotting_test.c@>=

@<Test program include files@>;

@<Show how to call test program@>;

int main(int argc, char *argv[])
{
    extern char *optarg;            /* Imported from |getopt| */
    extern int optind;

    int opt;

@<Process command-line options@>;@/

    return 0;
}

@
We use |getopt| to process command line options.  This
permits aggregation of options without arguments and
both \.{-d}{\it arg} and \.{-d} {\it arg} syntax.
@<Process command-line options@>=
    while ((opt = getopt(argc, argv, "nu-:")) != -1) {
        switch (opt) {

            case 'u':             /* \.{-u}  Print how-to-call information */
            case '?':
                usage();
                return 0;

            case '-':             /* \.{--}  Extended options */
                switch (optarg[0]) {
                    case 'c':     /* \.{--copyright} */
                        cout << "This program is in the public domain.\n";
                        return 0;

                    case 'h':     /* \.{--help} */
                        usage();
                        return 0;

                    case 'v':     /* \.{--version} */
                        cout << PRODUCT << " " << VERSION << "\n";
                        cout << "Last revised: " << REVDATE << "\n";
                        cout << "The latest version is always available\n";
                        cout << "at http://www.fourmilab.ch/eggtools/eggshell\n";
                        return 0;
                }
        }
    }

@ Procedure |usage|
prints how-to-call information.

@<Show how to call test program@>=
static void usage(void)
{
    cout << PRODUCT << "  --  Analyse eggsummary files.  Call:\n";
    cout << "              " << PRODUCT << " [options] [infile] [outfile]\n";
    cout << "\n";
    cout << "Options:\n";
    cout << "           --copyright       Print copyright information\n";
    cout << "           -u, --help        Print this message\n";
    cout << "           --version         Print version number\n";
    cout << "\n";
    cout << "by John Walker\n";
    cout << "http://www.fourmilab.ch/\n";
}

@
We need the following definitions to compile the test program.

@<Test program include files@>=

#include "config.h"         // Our configuration

/* C++ include files  */
#include <iostream>
#include <exception>
#include <stdexcept>
#include <string>
using namespace std;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef HAVE_GETOPT
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#else
#include "getopt.h"     /* No system \.{getopt}---use our own */
#endif
#include "plotting.h"        // Class definitions for this package


@** Index.
The following is a cross-reference table for \.{plotting}.
Single-character identifiers are not indexed, nor are
reserved words.  Underlined entries indicate where
an identifier was declared.
